syntax = "proto3";

package aea.valory.http.v1_0_0;

message HttpMessage{

  // Performatives and contents
  message Request_Performative{
    string method = 1;
    string url = 2;
    string version = 3;
    string headers = 4;
    bytes body = 5;
  }

  message Response_Performative{
    string version = 1;
    int32 status_code = 2;
    string status_text = 3;
    string headers = 4;
    bytes body = 5;
  }


  oneof performative{
    Request_Performative request = 5;
    Response_Performative response = 6;
  }
}
